#include "Camera.h"
#define _USE_MATH_DEFINES
#include <math.h>

Camera::Camera(const Eigen::Vector3f & pivot)
	: pivot(pivot), radius(10.f), azimuth(0.f), polar(static_cast<float>(M_PI / 6.)),
	screen_size(100.f, 100.f), opening_angle(static_cast<float>(51. / 180.*M_PI)),
	near_dist(.1f), far_dist(std::numeric_limits<float>::infinity())
{
	update();
}

void Camera::increaseAngles(const Eigen::Vector2f& incr)
{
	increaseAngles(incr[0], incr[1]);
}

void Camera::increaseAngles(float incr_azimuth, float incr_polar)
{
	azimuth += incr_azimuth;
	polar = std::max(std::min(polar + incr_polar, static_cast<float>(0.499*M_PI)),
		static_cast<float>(-0.499*M_PI));
	update(true, false);
}

void Camera::increasePivot(const Eigen::Vector3f& incr_pivot)
{
	pivot += incr_pivot;
	update(true, false);
}

void Camera::setScreenSize(const Eigen::Vector2f& size)
{
	screen_size = size;
	update(false, true);
}

void Camera::setRadius(float rad)
{
	radius = std::max(1.e-1f, rad);
	update(true, false);
}

void Camera::getCornerDirections(Eigen::Vector3f& ll, Eigen::Vector3f& lr, Eigen::Vector3f& ul, Eigen::Vector3f& ur) const
{
	auto r = near_dist / fracx;
	auto t = near_dist / fracy;
	auto mnz = -near_dist * dirz;
	auto rx = r * dirx;
	auto ty = t * diry;
	ll = mnz - rx - ty;
	lr = mnz + rx - ty;
	ul = mnz - rx + ty;
	ur = mnz + rx + ty;
}

const Eigen::Vector3f& Camera::getDirection(int i) const
{
	switch (i)
	{
	case 0:
		return dirx;
	case 1:
		return diry;
	case 2:
		return dirz;
	default:
		std::cout << "Error: Direction between 0 and 2!" << std::endl;
		return dirx;
	}
}

void Camera::update(bool view, bool proj)
{
	if (view)
	{
		position = pivot + Eigen::Vector3f(
			radius * cosf(polar) * cosf(azimuth),
			radius * sinf(polar),
			radius * cosf(polar) * sinf(azimuth));
		dirz = (position - pivot).normalized();
		dirx = Eigen::Vector3f(0.f, 1.f, 0.f).cross(dirz).normalized();
		diry = dirz.cross(dirx);

		Eigen::Matrix3f rot;
		rot.row(0) << dirx.transpose();
		rot.row(1) << diry.transpose();
		rot.row(2) << dirz.transpose();
		view_matrix.topLeftCorner<3, 3>() << rot;
		view_matrix.topRightCorner<3, 1>() << rot * (-position);
		view_matrix.row(3) << 0.f, 0.f, 0.f, 1.f;
	}

	if (proj)
	{
		auto frac = 1.f / tanf(opening_angle*0.5f);
		if (screen_size[0] > screen_size[1])
		{
			fracx = frac * screen_size[1] / screen_size[0];
			fracy = frac;
		}
		else
		{
			fracx = frac;
			fracy = frac * screen_size[0] / screen_size[1];
		}
		if (far_dist == std::numeric_limits<float>::infinity())
		{
			projection_matrix << fracx, 0.f, 0.f, 0.f,
				0.f, fracy, 0.f, 0.f,
				0.f, 0.f, -1.f, -2 * near_dist,
				0.f, 0.f, -1.f, 0.f;
		}
		else
		{
			projection_matrix << fracx, 0.f, 0.f, 0.f,
				0.f, fracy, 0.f, 0.f,
				0.f, 0.f, -(far_dist + near_dist) / (far_dist - near_dist), -2 * far_dist*near_dist / (far_dist - near_dist),
				0.f, 0.f, -1.f, 0.f;
		}
	}

	vp_matrix = projection_matrix * view_matrix;
}