#include "GLTexture.h"
#include "GLBuffer.h"

GLTexture1D::GLTexture1D(GLuint name, GLint internalFormat, GLsizei width, bool immutable, GLenum format, GLenum type)
	:
	name(name),
	internalFormat(internalFormat),
	width(width),
	minFilter(GL_LINEAR),
	magFilter(GL_LINEAR),
	wrap(GL_REPEAT),
	format(format),
	type(type)
{
	switch (format)
	{
	case GL_RED:
	case GL_GREEN:
	case GL_BLUE:
		num_channels = 1;
		break;
	case GL_RG:
		num_channels = 2;
		break;
	case GL_RGB:
		num_channels = 3;
		break;
	default:
		num_channels = 4;
		break;
	}
}

GLTexture1D::~GLTexture1D()
{
	glDeleteTextures(1, &name);
}

std::unique_ptr<GLTexture1D> GLTexture1D::newMutable(GLint internalFormat, GLsizei width, GLenum format, GLenum type, const void *data)
{
	GLuint name;
	glGenTextures(1, &name);
	GLTexture1D* ret = new GLTexture1D(name, internalFormat, width, false, format, type);
	glBindTexture(GL_TEXTURE_1D, name);
	glTexImage1D(GL_TEXTURE_1D, 0, internalFormat, width, 0, format, type, data);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, ret->getMinFilter());
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, ret->getMagFilter());
	return std::unique_ptr<GLTexture1D>(ret);
}

void GLTexture1D::setFilters(GLint minFilter, GLint magFilter)
{
	this->minFilter = minFilter;
	this->magFilter = magFilter;
	glBindTexture(GL_TEXTURE_1D, name);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, minFilter);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, magFilter);
}

void GLTexture1D::setWrap(GLint wrap)
{
	this->wrap = wrap;
	glBindTexture(GL_TEXTURE_1D, name);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, wrap);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_T, wrap);
}

void GLTexture1D::generateMipmaps() const
{
	glBindTexture(GL_TEXTURE_1D, name);
	glGenerateMipmap(GL_TEXTURE_1D);
}

void GLTexture1D::bindAndActive(int index) const
{
	glActiveTexture(GL_TEXTURE0 + index);
	glBindTexture(GL_TEXTURE_1D, name);
}




GLTexture2D::GLTexture2D(GLuint name, GLint internalFormat, GLsizei width, GLsizei height, bool immutable, GLenum format, GLenum type)
	:
	name(name),
	internalFormat(internalFormat),
	width(width),
	height(height),
	minFilter(GL_LINEAR),
	magFilter(GL_LINEAR),
	wrap(GL_REPEAT),
	format(format),
	type(type)
{
	switch (format)
	{
	case GL_RED:
	case GL_GREEN:
	case GL_BLUE:
		num_channels = 1;
		break;
	case GL_RG:
		num_channels = 2;
		break;
	case GL_RGB:
		num_channels = 3;
		break;
	default:
		num_channels = 4;
		break;
	}
}

GLTexture2D::~GLTexture2D()
{
	glDeleteTextures(1, &name);
}

std::unique_ptr<GLTexture2D> GLTexture2D::newMutable(GLint internalFormat, GLsizei width, GLsizei height, GLenum format, GLenum type, const void *data)
{
	GLuint name;
	glGenTextures(1, &name);
	GLTexture2D* ret = new GLTexture2D(name, internalFormat, width, height, false, format, type);
	glBindTexture(GL_TEXTURE_2D, name);
	glTexImage2D(GL_TEXTURE_2D, 0, internalFormat, width, height, 0, format, type, data);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, ret->getMinFilter());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, ret->getMagFilter());
	return std::unique_ptr<GLTexture2D>(ret);
}

void GLTexture2D::resize(GLsizei w, GLsizei h)
{
	width = w;
	height = h;
	glBindTexture(GL_TEXTURE_2D, name);
	glTexImage2D(GL_TEXTURE_2D, 0, internalFormat, w, h, 0, format, type, nullptr);
}

void GLTexture2D::setFilters(GLint minFilter, GLint magFilter)
{
	this->minFilter = minFilter;
	this->magFilter = magFilter;
	glBindTexture(GL_TEXTURE_2D, name);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minFilter);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, magFilter);
}

void GLTexture2D::setWrap(GLint wrap)
{
	this->wrap = wrap;
	glBindTexture(GL_TEXTURE_2D, name);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrap);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrap);
}

void GLTexture2D::generateMipmaps() const
{
	glBindTexture(GL_TEXTURE_2D, name);
	glGenerateMipmap(GL_TEXTURE_2D);
}

void GLTexture2D::bindAndActive(int index) const
{
	glActiveTexture(GL_TEXTURE0 + index);
	glBindTexture(GL_TEXTURE_2D, name);
}

void GLTexture2D::getImage(std::vector<unsigned char>& data, int& w, int& h, GLint level) const
{
	getImage(data, GL_UNSIGNED_BYTE, w, h, level);
}
void GLTexture2D::getImage(std::vector<float>& data, int& w, int& h, GLint level) const
{
	getImage(data, GL_FLOAT, w, h, level);
}

template<typename T>
void GLTexture2D::getImage(std::vector<T>& data, GLenum type, int& w, int& h, GLint level) const
{
	glBindTexture(GL_TEXTURE_2D, name);
	w = width / static_cast<GLsizei>(std::powf(2.f, static_cast<float>(level)));
	h = height / static_cast<GLsizei>(std::powf(2.f, static_cast<float>(level)));
	data.resize(std::max(1, w) * std::max(1, h) * num_channels);
	glGetTexImage(GL_TEXTURE_2D, level, format, type, data.data());
}

GLTextureCubemap::~GLTextureCubemap()
{
	glDeleteTextures(1, &name);
}

std::unique_ptr<GLTextureCubemap> GLTextureCubemap::newMutable(GLint internalFormat, GLsizei width, GLenum format, GLenum type)
{
	GLuint name;
	glGenTextures(1, &name);
	GLTextureCubemap* ret = new GLTextureCubemap(name, internalFormat, width, format, type, false);
	glBindTexture(GL_TEXTURE_CUBE_MAP, name);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, ret->getMinFilter());
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, ret->getMagFilter());
	return std::unique_ptr<GLTextureCubemap>(ret);
}

void GLTextureCubemap::setContents(int side, void* data)
{
	glBindTexture(GL_TEXTURE_CUBE_MAP, name);
	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + side, 0, internalFormat, width, width, 0, format, type, data);
}

void GLTextureCubemap::setFilters(GLint minFilter, GLint magFilter)
{
	this->minFilter = minFilter;
	this->magFilter = magFilter;
	glBindTexture(GL_TEXTURE_CUBE_MAP, name);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, minFilter);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, magFilter);
}
void GLTextureCubemap::generateMipmaps() const
{
	glBindTexture(GL_TEXTURE_CUBE_MAP, name);
	glGenerateMipmap(GL_TEXTURE_CUBE_MAP);
}
void GLTextureCubemap::bindAndActive(int index) const
{
	glActiveTexture(GL_TEXTURE0 + index);
	glBindTexture(GL_TEXTURE_CUBE_MAP, name);
}