#include <nanogui/nanogui.h>
#include "RippleApp.h"
#include "GLBuffer.h"
#include "GLShader.h"
#include "GLProgram.h"
#include "GLVertexArray.h"
#include "Images.h"
#include <ctime>

#include "MeshHelper.h"

RippleApp::RippleApp(const nanogui::Vector2i& size)
	: nanogui::Screen(size, "Ripples", true, false, 8, 8, 24, 8, 8, 4, 4), cam_angle_speed(0.01f, 0.01f), cam_scroll_speed(1.05f)
{
	srand(static_cast <unsigned> (time(0)));

	// Initialize OpenGL resources like the screen-space grid, textures storage, and framebuffers
	MeshOutput mo;
	num_samples_x = 10; num_samples_y = 10;
	MeshHelper::makeRegularGrid(mo, num_samples_x-1, num_samples_y-1);
	trigger_resize = true;
	water_height_texture = GLTexture2D::newMutable(GL_R32F, num_samples_x, num_samples_y, GL_RED, GL_FLOAT);
	water_height_framebuffer = std::make_unique<rgl::GLFramebuffer>();
	water_height_framebuffer->attach(*water_height_texture, GL_COLOR_ATTACHMENT0);
	water_height_framebuffer->setDrawBuffer(GL_COLOR_ATTACHMENT0);
	water_height_framebuffer->printFramebufferCompleteMessage();
	water_normal_texture = GLTexture2D::newMutable(GL_RGB32F, num_samples_x, num_samples_y, GL_RGB, GL_FLOAT);
	water_normal_framebuffer = std::make_unique<rgl::GLFramebuffer>();
	water_normal_framebuffer->attach(*water_normal_texture, GL_COLOR_ATTACHMENT0);
	water_normal_framebuffer->setDrawBuffer(GL_COLOR_ATTACHMENT0);
	water_normal_framebuffer->printFramebufferCompleteMessage();

	// Wave Renderer
	wqr = std::make_unique<WaterQuadRenderer>();
	wqr->setWaveTypes(&gravity_waves, &capillary_waves);

	rgl::GLFramebuffer::unbind();

	// More OpenGL Initialization
	grid_positions = GLBuffer<float>::makeMutable(GL_ARRAY_BUFFER, mo.positions, GL_STATIC_DRAW);
	grid_indices = GLBuffer<unsigned int>::makeMutable(GL_ELEMENT_ARRAY_BUFFER, mo.indices, GL_STATIC_DRAW);

	grid_vao = std::make_unique<GLVertexArray>();
	grid_vao->attachBuffer(*grid_positions, 0, 2);
	grid_vao->attachBuffer(*grid_positions, 1, 2);
	grid_vao->attachElementBuffer(*grid_indices);

	MeshHelper::makeRegularGrid(mo, 1, 1);
	quad_positions = GLBuffer<float>::makeMutable(GL_ARRAY_BUFFER, mo.positions, GL_STATIC_DRAW);
	quad_indices = GLBuffer<unsigned int>::makeMutable(GL_ELEMENT_ARRAY_BUFFER, mo.indices, GL_STATIC_DRAW);
	quad_vao = std::make_unique<GLVertexArray>();
	quad_vao->attachBuffer(*quad_positions, 0, 2);
	quad_vao->attachElementBuffer(*quad_indices);

	// Initialize shader program for wave rendering
	basic_program = GLProgram::makeFromFiles({
		ShaderFromFile(GL_VERTEX_SHADER, "shaders/basic.vs"),
		ShaderFromFile(GL_FRAGMENT_SHADER, "shaders/basic.fs")
		});
	basic_program->link();
	ul_basic_program.mvp_matrix = basic_program->getUniformLocation("mvp_matrix");
	ul_basic_program.tex = basic_program->getUniformLocation("tex");
	ul_basic_program.cam_position = basic_program->getUniformLocation("cam_position");
	ul_basic_program.cam_dir_ll = basic_program->getUniformLocation("cam_dir_ll");
	ul_basic_program.cam_dir_right = basic_program->getUniformLocation("cam_dir_right");
	ul_basic_program.cam_dir_up = basic_program->getUniformLocation("cam_dir_up");

	std::vector<unsigned char> tex_data;
	unsigned int tw, th;
	Images::loadPNG("images/grid_pattern_uv.png", tex_data, tw, th);
	test_texture = GLTexture2D::newMutable(GL_RGBA8, tw, th, GL_RGBA, GL_UNSIGNED_BYTE, tex_data.data());
	test_texture->generateMipmaps();
	test_texture->setFilters(GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR);

	// Load environment cubemap
	env_map = GLTextureCubemap::newMutable(GL_RGBA8, 1024, GL_RGBA, GL_UNSIGNED_BYTE);
	for (int i = 0; i < 6; i++)
	{
		tex_data.clear();
		Images::loadPNG("images/cube" + std::to_string(i) + ".png", tex_data, tw, th);
		env_map->setContents(i, tex_data.data());
	}
	env_map->generateMipmaps();
	env_map->setFilters(GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR);

	// Shader Program for envmap rendering
	envmap_program = GLProgram::makeFromFiles({
		ShaderFromFile(GL_VERTEX_SHADER, "shaders/envmap.vs"),
		ShaderFromFile(GL_FRAGMENT_SHADER, "shaders/envmap.fs")
		});
	envmap_program->link();
	ul_envmap_program.cam_dir_ll = envmap_program->getUniformLocation("cam_dir_ll");
	ul_envmap_program.cam_dir_right = envmap_program->getUniformLocation("cam_dir_right");
	ul_envmap_program.cam_dir_up = envmap_program->getUniformLocation("cam_dir_up");
	ul_envmap_program.env_map = envmap_program->getUniformLocation("env_map");

	// Init camera
	cam = std::make_unique<Camera>();
	cam->setRadius(5.f);
	cam2 = std::make_unique<Camera>(*cam);
	wqr->setCamera(cam2.get());

	// Create GUI
	gui = new nanogui::FormHelper(this);
	gui_window = gui->addWindow(Eigen::Vector2i(10, 10), "Simulation");
	gui->addGroup("Properties");
	drop_radius = 0.3f;
	//gui->addVariable("Drop Radius", drop_radius)->setFixedWidth(60);
	gui->addVariable("Speed", time_speed)->setFixedWidth(60);
	gravity_waves = true;
	capillary_waves = true;
	gui->addVariable("Gravity Waves", gravity_waves);
	gui->addVariable("Capillary Waves", capillary_waves);
	//gui->addVariable("Wireframe", wireframe);
	samples_per_pixel_x = .5f;
	samples_per_pixel_y = 1.f;
	//samples_per_pixel_x = .1f;
	//samples_per_pixel_y = .1f;
	gui->addVariable<float>("spp x",
		[this](float sppx) {samples_per_pixel_x = sppx; trigger_resize = true; },
		[this]() {return samples_per_pixel_x; })->setFixedWidth(60);
	gui->addVariable<float>("spp y",
		[this](float sppy) {samples_per_pixel_y = sppy; trigger_resize = true; },
		[this]() {return samples_per_pixel_y; })->setFixedWidth(60);
	gui->addVariable<int>("Raindrops/s",
		[this](int val) {raindrops_per_second = val; raindrops_fallen = 0; timestamp_rain_started = last_time_stamp; },
		[this]() {return raindrops_per_second; })->setFixedWidth(60);
	gui->addVariable("Draw Waves", draw_waves);
	gui->addVariable("Update Waves", update_waves);
	
	performLayout();

	last_time_stamp = glfwGetTime();

	glEnable(GL_TEXTURE_CUBE_MAP_SEAMLESS);
}

void RippleApp::printGLError()
{
	auto codes = std::array<int, 4>({
		GL_INVALID_ENUM,
		GL_INVALID_VALUE,
		GL_OUT_OF_MEMORY,
		GL_INVALID_FRAMEBUFFER_OPERATION});
	auto code_names = std::array<std::string, codes.size()>({
		"GL_INVALID_ENUM",
		"GL_INVALID_VALUE",
		"GL_OUT_OF_MEMORY",
		"GL_INVALID_FRAMEBUFFER_OPERATION"
		});

	auto err = glGetError();
	for (auto i = 0; i < codes.size(); i++)
	{
		if (err == codes[i])
			std::cout << code_names[i] << std::endl;
	}
}

void RippleApp::drawContents()
{
	auto t = glfwGetTime();
	auto dt = std::min(0.2, t - last_time_stamp);

	{
		// Camera handling
		auto up = static_cast<float>(r_down) - static_cast<float>(f_down);
		auto back = static_cast<float>(s_down) - static_cast<float>(w_down);
		auto right = static_cast<float>(d_down) - static_cast<float>(a_down);
		Eigen::Vector3f back_dir = cam->getDirection(2);
		back_dir[1] = 0.0f;
		back_dir.normalize();
		Eigen::Vector3f right_dir = cam->getDirection(0);
		right_dir[1] = 0.0f;
		right_dir.normalize();
		Eigen::Vector3f up_dir(0.f, 1.f, 0.f);

		cam->increasePivot(3.f * static_cast<float>(dt) * (up*up_dir + back * back_dir + right * right_dir));
	}

	glEnable(GL_DEPTH_TEST);

	//glClearColor(1.f, 0.f, 0.f, 1.f);
	glClear(GL_DEPTH_BUFFER_BIT);
	glDisable(GL_BLEND);
	glDepthFunc(GL_LEQUAL);
	glBlendEquation(GL_FUNC_ADD);
	glBlendFunc(GL_ONE, GL_ONE);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);

	if (trigger_resize)
	{
		// Resize opengl resources with new size
		cam->setScreenSize(screen_size);

		MeshOutput mo;
		num_samples_x = static_cast<int>(screen_size[0] * samples_per_pixel_x) + 1;
		num_samples_y = static_cast<int>(screen_size[1] * samples_per_pixel_y) + 1;
		MeshHelper::makeRegularGrid(mo, num_samples_x-1, num_samples_y-1);
		grid_positions->uploadData(mo.positions, true);
		grid_indices->uploadData(mo.indices, true);

		water_height_texture->resize(num_samples_x, num_samples_y);
		water_normal_texture->resize(num_samples_x, num_samples_y);

		trigger_resize = false;

		std::cout << "Resized: " << screen_size[0] << " " << screen_size[1] << std::endl;
	}

	if (!camera_fixed)
		*cam2 = *cam;

	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	water_height_framebuffer->bind();
	glViewport(0, 0, num_samples_x, num_samples_y);
	float zeros[] = { 0.f,0.f,0.f,0.f };
	glClearBufferfv(GL_COLOR, 0, zeros);

	if (rain_enabled)
	{
		int raindrops_target = static_cast<int>(std::ceil(static_cast<long double>((last_time_stamp - timestamp_rain_started) * raindrops_per_second)));
		int raindrops_now = std::max(0, raindrops_target - raindrops_fallen);
		for (int i = 0; i < raindrops_now; i++)
		{
			Eigen::Vector2i pos = ((Eigen::Array2d::Random()*0.5 + 0.5) * screen_size.cast<double>().array()).round().cast<int>().matrix();
			addRipple(pos);
		}
		raindrops_fallen += raindrops_now;
	}

	// Pass 1: Compute Wave functions in compute shader
	if (update_waves)
		wqr->update(dt * time_speed);
	// Pass 2: Store wave heights in screen-space texture
	if (draw_waves)
		wqr->drawWaves();

	//std::vector<float> imgdata;
	//int w, h;
	//water_height_texture->getImage(imgdata, w, h);

	rgl::GLFramebuffer::unbind();
	grid_vao->bind();

	glDisable(GL_BLEND);
	glViewport(0, 0, static_cast<GLsizei>(screen_size[0]), static_cast<GLsizei>(screen_size[1]));

	// Pass 3: Draw projected grid with wave height displacement
	basic_program->use();
	water_height_texture->bindAndActive(0);
	env_map->bindAndActive(1);
	glUniform1i(ul_basic_program.tex, 0);
	glUniformMatrix4fv(ul_basic_program.mvp_matrix, 1, GL_FALSE, cam->getVPMatrix().data());
	glUniform1i(ul_basic_program.env_map, 1);

	auto cam_pos = cam2->getPosition();
	Eigen::Vector3f ll, lr, ul, ur;
	cam2->getCornerDirections(ll, lr, ul, ur);
	Eigen::Vector3f right = lr - ll;
	Eigen::Vector3f up = ul - ll;
	glUniform4f(ul_basic_program.cam_position, cam_pos[0], cam_pos[1], cam_pos[2], cam_pos[1] > 0.f ? -1.f : 1.f);
	glUniform3fv(ul_basic_program.cam_dir_ll, 1, ll.data());
	glUniform3fv(ul_basic_program.cam_dir_right, 1, right.data());
	glUniform3fv(ul_basic_program.cam_dir_up, 1, up.data());

	if (wireframe)
	{
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	}
	else
	{
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}
	glDrawElements(GL_TRIANGLES, grid_indices->getNumElements(), grid_indices->getGLType(), nullptr);

	// Draw envmap in background
	GLVertexArray::unbind();
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	envmap_program->use();
	env_map->bindAndActive(0);
	glUniform1i(ul_envmap_program.env_map, 0);
	glUniform3fv(ul_envmap_program.cam_dir_ll, 1, ll.data());
	glUniform3fv(ul_envmap_program.cam_dir_right, 1, right.data());
	glUniform3fv(ul_envmap_program.cam_dir_up, 1, up.data());

	quad_vao->bind();
	glDrawElements(GL_TRIANGLES, quad_indices->getNumElements(), quad_indices->getGLType(), nullptr);

	//glBindFramebuffer(GL_READ_FRAMEBUFFER, water_height_framebuffer->getName());
	//glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
	//glBlitFramebuffer(0, 0, water_height_texture->getWidth(), water_height_texture->getHeight(), 0, 0, water_height_texture->getWidth(), water_height_texture->getHeight(), GL_COLOR_BUFFER_BIT, GL_LINEAR);

	printGLError();

	last_time_stamp += dt;
}

bool RippleApp::mouseButtonEvent(const Eigen::Vector2i &p, int button, bool down, int modifiers)
{
	// Left MB: camera handling
	// Right MB: add wave
	auto ret = Screen::mouseButtonEvent(p, button, down, modifiers);

	if (down && button == GLFW_MOUSE_BUTTON_LEFT && !ret)
		left_pressed = true;
	if (left_pressed && button == GLFW_MOUSE_BUTTON_LEFT && !down)
		left_pressed = false;
	if (down && button == GLFW_MOUSE_BUTTON_RIGHT && !ret)
		addRipple(p);

	return ret;
}

void RippleApp::addRipple(const Eigen::Vector2i &sp)
{
	// Find point where mouse ray intersects water plane, add wave there
	Eigen::Vector2f frac = (pixelRatio() * sp.cast<float>()).cwiseQuotient(screen_size);
	frac[1] = 1.f - frac[1];
	Eigen::Vector3f ll, lr, ul, ur;
	cam->getCornerDirections(ll, lr, ul, ur);
	auto p = cam->getPosition();
	auto right = lr - ll;
	auto up = ul - ll;
	auto dir = ll + frac[0] * right + frac[1] * up;
	if (dir[1] > -1e-3)
		return;

	auto q = p - (p[1] / dir[1])*dir;
	wqr->addWave(Eigen::Vector2f(q[2],q[0]));
}

bool RippleApp::keyboardEvent(int key, int scancode, int action, int modifier)
{
	// Store relevant keyboard events for camera handling
	if (key == GLFW_KEY_LEFT_CONTROL)
		ctrl_down = action;
	else if (key == GLFW_KEY_C && action == 1)
		camera_fixed = !camera_fixed;
	else if (key == GLFW_KEY_W)
		w_down = action;
	else if (key == GLFW_KEY_A)
		a_down = action;
	else if (key == GLFW_KEY_S)
		s_down = action;
	else if (key == GLFW_KEY_D)
		d_down = action;
	else if (key == GLFW_KEY_R)
		r_down = action;
	else if (key == GLFW_KEY_F)
		f_down = action;
	else if (key == GLFW_KEY_B && action == GLFW_PRESS)
	{
		rain_enabled = !rain_enabled;
		raindrops_fallen = 0;
		timestamp_rain_started = last_time_stamp;
	}

	return Screen::keyboardEvent(key, scancode, action, modifier);
}

bool RippleApp::resizeEvent(const Eigen::Vector2i & s)
{
	// Trigger resizing of OpenGL resources in next update
	auto ret = Screen::resizeEvent(s);
	screen_size = s.cast<float>() * pixelRatio();
	trigger_resize = true;

	return ret;
}

bool RippleApp::mouseMotionEvent(const Eigen::Vector2i &p, const Eigen::Vector2i &rel, int button, int modifiers)
{
	// Camera Handling
	auto ret = Screen::mouseMotionEvent(p, rel, button, modifiers);
	if (left_pressed)
	{
		auto mult = ctrl_down ? 0.2f : 1.f;
		cam->increaseAngles(rel.cast<float>().cwiseProduct(mult * cam_angle_speed));
		// std::cout << cam->getPosition() << std::endl << std::endl;
	}
	
	return ret;
}

bool RippleApp::scrollEvent(const Eigen::Vector2i &p, const Eigen::Vector2f &rel)
{
	// Camera Handling
	auto ret = Screen::scrollEvent(p, rel);
	if (!ret)
	{
		auto mult = ctrl_down ? 0.2f : 1.f;
		cam->setRadius(cam->getRadius() * std::powf(cam_scroll_speed, -rel[1] * mult));
	}
	return ret;
}