#include "WaterQuadRenderer.h"

WaterQuadRenderer::WaterQuadRenderer()
{
	// OpenGL Buffer with quad geometry (used as wave bounding box)
	wc = std::make_unique<WaveComputations>();
	Eigen::Matrix<float, 2, 4> pos;
	pos <<
		-.5f, .5f, .5f, -.5f,
		-.5f, -.5f, .5f, .5f;
	quad_positions = GLBuffer<float>::makeMutable(GL_ARRAY_BUFFER, pos, GL_STATIC_DRAW);
	Eigen::Matrix<unsigned, 3, 2> ind;
	ind <<
		0u, 0u,
		1u, 2u,
		2u, 3u;
	quad_indices = GLBuffer<unsigned>::makeMutable(GL_ELEMENT_ARRAY_BUFFER, ind, GL_STATIC_DRAW);
	quad_translations = GLBuffer<float>::makeMutable(GL_ARRAY_BUFFER, 1, GL_DYNAMIC_DRAW);
	quad_vao = std::make_unique<GLVertexArray>();
	quad_vao->attachBuffer(*quad_positions, 0u, 2);
	quad_vao->attachBuffer(*quad_translations, 1u, 2, 1u);
	quad_vao->attachBuffer(*wc->getWaveDataBuffer(), 2u, 3, 1u);
	quad_vao->attachElementBuffer(*quad_indices);

	// Shader that renderer wave height into screen-space texture
	wave_program = GLProgram::makeFromFiles({
		ShaderFromFile(GL_VERTEX_SHADER, "shaders/waves.vs"),
		ShaderFromFile(GL_FRAGMENT_SHADER, "shaders/waves.fs")
		});
	wave_program->link();
	ul_wave_program.mvp_matrix = wave_program->getUniformLocation("mvp_matrix");
	ul_wave_program.R_star = wave_program->getUniformLocation("R_star");
	ul_wave_program.wave_height_texture = wave_program->getUniformLocation("wave_height_texture");

	wave_quads.resize(1);
}

void WaterQuadRenderer::addWave(const Eigen::Vector2f& pos)
{
	wave_quads.push_back(WaveQuad(pos));
}

void WaterQuadRenderer::update(double dt)
{
	//dt *= 0.1;
	timestamp += dt;
	//if (static_cast<int>(timestamp / 5.0) > wave_counter)
	//{
	//	wave_counter++;
	//	wave_quads.push_back(WaveQuad());
	//}

	// Update each wave with timestep
	auto iter = wave_quads.begin();
	while (iter != wave_quads.end())
	{
		iter->update(dt);
		if (iter->isFinished())
		{
			wave_quads.erase(iter++);
		}
		else
		{
			iter++;
		}
	}

	// Compute wave heights in compute shader
	wc->computeWaves(wave_quads, *gravity_waves, *capillary_waves);

	// Update positions of wave rings
	Eigen::Matrix<float, 2, -1> translations(2, wave_quads.size());
	int i = 0;
	for (auto &w : wave_quads)
	{
		translations.col(i++) << w.getPosition();
	}
	quad_translations->uploadData(translations);
}

void WaterQuadRenderer::drawWaves()
{
	// Draw wave height onto screen-space texture
	wave_program->use();
	glUniformMatrix4fv(ul_wave_program.mvp_matrix, 1, GL_FALSE, cam->getVPMatrix().data());
	glUniform1f(ul_wave_program.R_star, static_cast<float>(wc->getRStar()));
	glUniform1i(ul_wave_program.wave_height_texture, 0);
	wc->getWaveHeightTexture()->bindAndActive(0);

	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glEnable(GL_BLEND);
	//glDisable(GL_CULL_FACE);
	quad_vao->bind();
	glDrawElementsInstanced(GL_TRIANGLES, quad_indices->getNumElements(), quad_indices->getGLType(), nullptr, static_cast<GLsizei>(wave_quads.size()));
}