#include "WaveComputations.h"
#include <cmath>
#include "MathHelper.h"

#define WC_PI (3.14159265359)

WaveComputations::WaveComputations()
{
	// Init Constants (for unit conversion)
	tau = tau_star / (rho * g * R_star * R_star);
	W = It * 3 / (4 * WC_PI);
	W_star = W * sqrt(g * R_star);
	alpha = (3 * sqrt(g*R_star)) / (16 * R_star*W_star);
	beta = (3 * g) / (2 * R_star*R_star*R_star * W_star);
	precomputeDominantWavenumbers();

	// Compute Shader Program for wave computation
	compute_wave_program = GLProgram::makeFromFiles({
		ShaderFromFile(GL_COMPUTE_SHADER, "shaders/compute_wave.comp")
		});
	compute_wave_program->link();

	ul_compute_wave_program.global_size = compute_wave_program->getUniformLocation("global_size");
	ul_compute_wave_program.V_bounds = compute_wave_program->getUniformLocation("V_bounds");

	ul_compute_wave_program.g = compute_wave_program->getUniformLocation("g");
	ul_compute_wave_program.R_star = compute_wave_program->getUniformLocation("R_star");
	ul_compute_wave_program.nu = compute_wave_program->getUniformLocation("nu");
	ul_compute_wave_program.tau = compute_wave_program->getUniformLocation("tau");
	ul_compute_wave_program.alpha = compute_wave_program->getUniformLocation("alpha");
	ul_compute_wave_program.wave_types = compute_wave_program->getUniformLocation("wave_types");
	

	ul_compute_wave_program.k_texture = compute_wave_program->getUniformLocation("k_texture");
	ul_compute_wave_program.dV_texture = compute_wave_program->getUniformLocation("dV_texture");

	// Storage of per-wave parameters
	per_wave_data_buffer = GLBuffer<float>::makeMutable(GL_SHADER_STORAGE_BUFFER, 3 * max_num_waves, GL_DYNAMIC_DRAW);
}


double WaveComputations::sigma(double k)
{
	return sqrt(k * (1. + tau * k*k));
}

double WaveComputations::V(double k)
{
	return (1. + 3.*tau*k*k) / (2. * sqrt(k*(1. + tau * k*k)));
}

double WaveComputations::dV(double k)
{
	auto ksq = k * k;
	auto tmp = tau * ksq * k + k;
	return (3.*tau*ksq*(tau*ksq + 2) - 1.) / (4. * sqrt(tmp*tmp*tmp));
}

double WaveComputations::ddV(double k)
{
	auto ksq = k * k;
	return -(3. * (ksq*tau - 1.) * (ksq*tau*(ksq*tau + 6) + 1)) / (8 * sqrt(pow(ksq*k*tau + k, 5.)));
}

void WaveComputations::computeWaves(const std::list<WaveQuad>& waves, bool gw, bool cw)
{
	// Update per-wave parameters
	Eigen::Matrix<float, 3, -1> wave_data(3, max_num_waves);
	computeWaveData(wave_data, waves);
	per_wave_data_buffer->uploadData(wave_data);

	glClearTexImage(wave_height_texture->name, 0, GL_RED, GL_FLOAT, nullptr);

	// Compute waves and store result as rows of a texture
	compute_wave_program->use();
	glUniform1i(ul_compute_wave_program.k_texture, 0);
	glUniform1i(ul_compute_wave_program.dV_texture, 1);
	bindTextures(0, 1, 0);
	glUniform2ui(ul_compute_wave_program.global_size, num_wave_samples, static_cast<GLuint>(waves.size()));
	glUniform2f(ul_compute_wave_program.V_bounds, static_cast<GLfloat>(V_min), static_cast<GLfloat>(V_max));
	glUniform1f(ul_compute_wave_program.g, static_cast<GLfloat>(g));
	glUniform1f(ul_compute_wave_program.R_star, static_cast<GLfloat>(R_star));
	glUniform1f(ul_compute_wave_program.nu, static_cast<GLfloat>(nu));
	glUniform1f(ul_compute_wave_program.tau, static_cast<GLfloat>(tau));
	glUniform1f(ul_compute_wave_program.alpha, static_cast<GLfloat>(alpha));
	glUniform2f(ul_compute_wave_program.wave_types, static_cast<float>(gw), static_cast<float>(cw));
	glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, per_wave_data_buffer->getName());

	glDispatchCompute(num_wave_samples / 512u, static_cast<GLuint>(waves.size()), 1u);

	glMemoryBarrier(GL_ALL_BARRIER_BITS);
}

void WaveComputations::computeWaveData(Eigen::Matrix<float, 3, -1>& data, const std::list<WaveQuad>& waves) const
{
	// Update per-wave data such as time, min/max radius
	int i = 0;
	auto fac = sqrt(g / R_star);
	for (auto const &wave : waves)
	{
		auto t_star = wave.getTime();
		auto t_star_eff = (t_star*t_star + t_star + 0.15) / (t_star + 1.);
		auto t = t_star * fac;
		auto t_eff = t_star_eff * fac;
		data.col(i++) << static_cast<float>(t * V_min), static_cast<float>(t*V_min*4.f), static_cast<float>(t_eff);
	}
}

void WaveComputations::precomputeDominantWavenumbers()
{
	// Precompute lookup table for dominant wavenumbers as a function of group speed
	// Store result in red and green channel of a texture
	// In a second texture, store group-speed derivatives at the dominant wavenumbers
	auto lambda_V = [this](double k) {return V(k); };
	auto lambda_dV = [this](double k) {return dV(k); };
	auto lambda_ddV = [this](double k) {return ddV(k); };
	auto k_mid = MathHelper::newtonsMethod(lambda_dV, lambda_ddV, 0., 0.25);
	V_max = 100.0;
	V_min = V(k_mid);
	auto k_lb = MathHelper::newtonsMethod(lambda_V, lambda_dV, V_max, 1.e-5);
	auto k_ub = MathHelper::newtonsMethod(lambda_V, lambda_dV, V_max, 1.e3);

	auto n = 2048u;
	double interv = (V_max - V_min) / static_cast<double>(n);
	double V_start = V_min + interv * 0.5;
	
	V_vals = Eigen::VectorXf(n, 1);
	k_vals = Eigen::Matrix2Xf(2, n);
	dV_vals = Eigen::Matrix2Xf(2, n);
	for (unsigned i = 0; i < n; i++)
	{
		auto V_val = V_start + interv * i;
		auto k1 = MathHelper::newtonsMethod(lambda_V, lambda_dV, V_val, k_lb);
		auto frac = (1. + static_cast<float>(i)) / n;
		auto k2 = MathHelper::newtonsMethod(lambda_V, lambda_dV, V_val, 0.01*frac*k_ub+(1.- 0.01*frac)*k_mid);
		V_vals(i) = static_cast<float>(V_val);
		k_vals.col(i) << static_cast<float>(k1), static_cast<float>(k2);
		dV_vals.col(i) << static_cast<float>(dV(k1)), static_cast<float>(dV(k2));
	}

	k_texture = GLTexture1D::newMutable(GL_RG32F, n, GL_RG, GL_FLOAT, k_vals.data());
	k_texture->setWrap(GL_CLAMP_TO_EDGE);
	dV_texture = GLTexture1D::newMutable(GL_RG32F, n, GL_RG, GL_FLOAT, dV_vals.data());
	dV_texture->setWrap(GL_CLAMP_TO_EDGE);

	wave_height_texture = GLTexture2D::newMutable(GL_R32F, num_wave_samples, max_num_waves, GL_RED, GL_FLOAT, nullptr);
	wave_height_texture->setWrap(GL_CLAMP_TO_BORDER);
	glBindTexture(GL_TEXTURE_2D, wave_height_texture->name);
	GLfloat col[] = { 0.f, 0.f, 0.f, 1.f };
	glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, col);
}

void WaveComputations::bindTextures(int k, int dV, int water_height) const
{
	// Bind lookup tables for wavenumber and group speed derivatives
	k_texture->bindAndActive(k);
	dV_texture->bindAndActive(dV);
	// Bind screen-space wave height texture
	glBindImageTexture(water_height, wave_height_texture->name, 0, GL_FALSE, 0, GL_WRITE_ONLY, GL_R32F);
}